//##############################################################################################################//
/*
    File-echo example.  

    A user thread pool is used in this example.
    Long operations, such as accessing a database, can be performed
    in a user pool in order not to block worker threads for a long time.
    File I/O is used as an example of long operations.


    CLIENT            SERVER
    
    file.ext -------> file.ext
                          |
    file.ext.echo <-------+


    File sending format:

    ----------------------
    | SIZE | NAME | FILE |
    ----------------------
    ^             ^
    |   HEADER    |

    SIZE (4 bytes)        - Data size ( 4 + MAX_PATH + sizeof( FILE ) ).
    NAME (MAX_PATH bytes) - File name.
    FILE (N bytes)        - File.

*/
//##############################################################################################################//
#include "CmnHdr.h"

#include <windows.h>
#include <windowsX.h>
#include <tchar.h>

#include "CPServer.h"

#include "resource.h"

//##############################################################################################################//
#define HEADER_SIZE     ( 4 + MAX_PATH )

// Messages to the user pool.
#define MSG_RECV        1   // Save data to the file.
#define MSG_SEND        2   // Read data from the file and send.

//##############################################################################################################//
typedef struct _CONNECTION_DATA{
    BYTE   bBuffer[BUF_SIZE];
    DWORD  dwBytesTransferred;
    DWORD  dwDataSize;
    DWORD  dwIndex;              // Index in the buffer.
    CHAR   szFileName[MAX_PATH];
    HANDLE hFile;
}CONNECTION_DATA, *PCONNECTION_DATA;

//##############################################################################################################//
CPServer PServer; // ProstoServer class.

HWND g_hwndMain                 = NULL;
BOOL g_fStarted                 = FALSE;
LONG g_lOnConnectCount          = 0;
LONG g_lOnDisconnectCount       = 0;
LONG g_lOnRemoveConnectionCount = 0;

//##############################################################################################################//
//##############################################################################################################//
//##############################################################################################################//
VOID LogError( PSTR pszFunc, DWORD dwError )
{
    HANDLE hFile = CreateFile( TEXT("FileEchoServer.err"), GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, 
						       OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if( INVALID_HANDLE_VALUE == hFile ){ return; }

    SetFilePointer( hFile, 0, NULL, FILE_END );

    CHAR szError[64] = {0};
    wsprintfA( szError, "%s: %d\r\n", pszFunc, dwError );

    DWORD dwWrite = 0;
    WriteFile( hFile, szError, lstrlenA( szError ), &dwWrite, NULL );

    CloseHandle( hFile );
}

//###############################################################################################################//
VOID CALLBACK UserPoolHandler( DWORD dwMessage, PCONNECTION pConnection, PIO_DATA pIoData )
{
    BOOL  fContinue = FALSE;
    DWORD dwSize    = 0;     // Saved/read size.

    __try{
        PCONNECTION_DATA pConnectionData = (PCONNECTION_DATA) pConnection->pvPerConnectionUserData;

        switch( dwMessage ){
            case MSG_RECV:
                // First receipt - creating a file.
                if( INVALID_HANDLE_VALUE == pConnectionData->hFile ){
                    lstrcpyA( pConnectionData->szFileName, (PSTR) &pConnectionData->bBuffer[4] );

                    pConnectionData->hFile = CreateFileA( pConnectionData->szFileName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL,
                                                          CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
                    if( INVALID_HANDLE_VALUE == pConnectionData->hFile ){ __leave; }

                    // If part of the file has been received.
                    if( pConnectionData->dwBytesTransferred > HEADER_SIZE ){
                        WriteFile( pConnectionData->hFile,
                                   &pConnectionData->bBuffer[HEADER_SIZE],
                                   pConnectionData->dwBytesTransferred - HEADER_SIZE, &dwSize, NULL );
                        }
                    }
                else{
                    // Otherwise writing to the file everything that has been received.
                    WriteFile( pConnectionData->hFile,
                               (PVOID) pConnectionData->bBuffer,
                               pIoData->dwBytesTransferred, &dwSize, NULL );
                    }

                // If not all data has been received.
                if( pConnectionData->dwBytesTransferred < pConnectionData->dwDataSize ){
                    if( PServer.PostRecv( pConnection,
                                          pConnectionData->bBuffer,
                                          BUF_SIZE,
                                          NULL ) != ERROR_SUCCESS ){ __leave; }
                    }
                else{
                    // The file is received, preparing and sending information about the file.
                    pConnectionData->dwBytesTransferred = 0;

                    CopyMemory( pConnectionData->bBuffer,
                                &pConnectionData->dwDataSize,
                                sizeof( pConnectionData->dwDataSize ) );

                    lstrcpyA( (PSTR) &pConnectionData->bBuffer[4], pConnectionData->szFileName );

                    // Setting the pointer to the beginning of the received file.
                    SetFilePointer( pConnectionData->hFile, 0, NULL, FILE_BEGIN );

                    // Sending the header.
                    if( PServer.PostSend( pConnection,
                                          pConnectionData->bBuffer,
                                          HEADER_SIZE,
                                          // Specifying the size being sent in the pvPerIoUserData
                                          // parameter in order to know it in OnSend.
                                          (PVOID) HEADER_SIZE ) != ERROR_SUCCESS ){ __leave; }
                    }
                break;

            case MSG_SEND:
                // Reading and sending.
                if( ReadFile( pConnectionData->hFile, pConnectionData->bBuffer, BUF_SIZE, &dwSize, NULL ) && dwSize ){
                    if( PServer.PostSend( pConnection,
                                          pConnectionData->bBuffer,
                                          dwSize,
                                          (PVOID) dwSize ) != ERROR_SUCCESS ){ __leave; }
                    }
                else{
                    // The file has been sent (or a read error occurred) - closing the connection.
                    __leave;
                    }
                break;
            }

        fContinue = TRUE;
        }
    __finally{
        if( !fContinue ){
            PServer.CloseConnection( pConnection );
            }
        }
}

//###############################################################################################################//
BOOL CALLBACK OnConnect( PCONNECTION pConnection )
{
    InterlockedIncrement( &g_lOnConnectCount );

    // Allocate connection data buffer.
    pConnection->pvPerConnectionUserData = (PCONNECTION_DATA) malloc( sizeof( CONNECTION_DATA ) );
    if( NULL == pConnection->pvPerConnectionUserData ){
        LogError( "malloc()", 0 );
        return FALSE;
        }

    PCONNECTION_DATA pConnectionData = (PCONNECTION_DATA) pConnection->pvPerConnectionUserData;

    pConnectionData->dwBytesTransferred = 0;
    pConnectionData->dwDataSize         = 0;
    pConnectionData->dwIndex            = 0;
    pConnectionData->hFile              = INVALID_HANDLE_VALUE;

    // Receive request.
    if( PServer.PostRecv( pConnection, 
                          pConnectionData->bBuffer,
                          BUF_SIZE,
                          NULL ) != ERROR_SUCCESS ){ return FALSE; }

    return TRUE;
}

//##############################################################################################################//
VOID CALLBACK OnDisconnect( PCONNECTION pConnection )
{
    InterlockedIncrement( &g_lOnDisconnectCount );

    PCONNECTION_DATA pConnectionData = (PCONNECTION_DATA) pConnection->pvPerConnectionUserData;

    if( pConnectionData->hFile != INVALID_HANDLE_VALUE ){
        CloseHandle( pConnectionData->hFile );
        }
}

//##############################################################################################################//
VOID CALLBACK OnRemoveConnection( PCONNECTION pConnection )
{
    InterlockedIncrement( &g_lOnRemoveConnectionCount );

    // Free memory.
    if( pConnection->pvPerConnectionUserData != NULL ){
        free( pConnection->pvPerConnectionUserData );
        }
}

//##############################################################################################################//
BOOL CALLBACK OnRecv( PCONNECTION pConnection, PIO_DATA pIoData )
{
    BOOL fContinue = FALSE;

    __try{
        PCONNECTION_DATA pConnectionData = (PCONNECTION_DATA) pConnection->pvPerConnectionUserData;

        pConnectionData->dwBytesTransferred += pIoData->dwBytesTransferred;

        // First receipt.
        if( 0 == pConnectionData->dwDataSize ){
            // Bad packet.
            if( pIoData->dwBytesTransferred < 4 ){ __leave; }

            // Getting the data size.
            CopyMemory( &pConnectionData->dwDataSize,
                        pConnectionData->bBuffer,
                        sizeof( pConnectionData->dwDataSize ) );

            // Bad packet.
            if( pConnectionData->dwDataSize < HEADER_SIZE ){ __leave; }
            }

        // If not the entire header has been received.
        if( pConnectionData->dwBytesTransferred < HEADER_SIZE ){
            if( PServer.PostRecv( pConnection,
                                  &pConnectionData->bBuffer[pConnectionData->dwBytesTransferred],
                                  BUF_SIZE - pConnectionData->dwBytesTransferred,
                                  NULL ) != ERROR_SUCCESS ){ __leave; }
            fContinue = TRUE;
            __leave;
            }

        // Save data to the file.
        DWORD dwRet = PServer.PostMessageToUserPool( MSG_RECV, pConnection, pIoData );
        if( dwRet != ERROR_SUCCESS ){
            LogError( "PostMessageToUserPool()", dwRet );
            __leave;
            }

        fContinue = TRUE;
        }
    __finally{}

    return fContinue;
}

//##############################################################################################################//
BOOL CALLBACK OnSend( PCONNECTION pConnection, PIO_DATA pIoData )
{
    BOOL fContinue = FALSE;

    __try{
        PCONNECTION_DATA pConnectionData = (PCONNECTION_DATA) pConnection->pvPerConnectionUserData;

        pConnectionData->dwBytesTransferred += pIoData->dwBytesTransferred;

        // If not all data has been sent.
        if( pIoData->dwBytesTransferred < (DWORD) pIoData->pvPerIoUserData ){
            DWORD dwLeft = (DWORD) pIoData->pvPerIoUserData;

            dwLeft                   -= pIoData->dwBytesTransferred;
            pConnectionData->dwIndex += pIoData->dwBytesTransferred;

            if( PServer.PostSend( pConnection,
                                  &pConnectionData->bBuffer[pConnectionData->dwIndex],
                                  dwLeft,
                                  (PVOID) dwLeft ) != ERROR_SUCCESS ){ __leave; }
            fContinue = TRUE;
            __leave;
            }

        pConnectionData->dwIndex = 0;

        // Read data from the file and send.
        DWORD dwRet = PServer.PostMessageToUserPool( MSG_SEND, pConnection, pIoData );
        if( dwRet != ERROR_SUCCESS ){
            LogError( "PostMessageToUserPool()", dwRet );
            __leave;
            }

        fContinue = TRUE;
        }
    __finally{}

    return fContinue;
}

//##############################################################################################################//
VOID CALLBACK OnError( PCWSTR pszFunc, DWORD dwError )
{
    TCHAR szError[64] = {0};
    wsprintf( szError, TEXT("%s(): %d"), pszFunc, dwError );
    SetDlgItemText( g_hwndMain, ID_STATIC_STATUS, szError );
}

//###############################################################################################################//
//###############################################################################################################//
//###############################################################################################################//
void DlgMain_OnCommand( HWND hWnd, int id, HWND hWndCtl, UINT codeNotify )
{
	switch( id ){
        case ID_BUTTON_START:{
            SetDlgItemText( hWnd, ID_STATIC_STATUS, TEXT("Starting...") );

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_START ), FALSE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_PORT    ), FALSE );

            g_lOnConnectCount          = 0;
            g_lOnDisconnectCount       = 0;
            g_lOnRemoveConnectionCount = 0;

            PSERVERPARAM param = {0};

            /*
            // Available in the full version.
            SYSTEM_INFO  sysinfo = {0};
            GetSystemInfo( &sysinfo );
            param.dwWorkerThreadsMin = sysinfo.dwNumberOfProcessors * 2;
            param.dwWorkerThreadsMax = sysinfo.dwNumberOfProcessors * 4;
            param.dwUserThreadsMin   = sysinfo.dwNumberOfProcessors * 4;
            param.dwUserThreadsMax   = sysinfo.dwNumberOfProcessors * 8;
            */

            param.dwWorkerThreadsIdleTimeout   = 60000;
            param.dwUserThreadsMin             = 1;
            param.dwUserThreadsMax             = 1;
            param.dwUserThreadsIdleTimeout     = 60000;
            param.dwCpuUsageMin                = 70;
            param.dwCpuUsageMax                = 90;
            param.dwMultipleIoReqsMax          = 1;
            param.dwPort                       = GetDlgItemInt( hWnd, ID_EDIT_PORT, NULL, FALSE );
            param.dwInterface                  = INADDR_ANY;
            param.dwSocketRecvBufferSize       = -1;
            param.dwSocketSendBufferSize       = -1;
            param.dwConnectionsIdleTimeout     = 500;
            param.dwInitAccepts                = 200;
            param.CALLBACKS.OnConnect          = OnConnect;
            param.CALLBACKS.OnDisconnect       = OnDisconnect;
            param.CALLBACKS.OnRemoveConnection = OnRemoveConnection;
            param.CALLBACKS.OnRecv             = OnRecv;
            param.CALLBACKS.OnSend             = OnSend;
            param.CALLBACKS.OnError            = OnError;
            param.CALLBACKS.UserPoolHandler    = UserPoolHandler;

            DWORD dwRet = PServer.Start( &param );
            if( dwRet != ERROR_SUCCESS ){
                TCHAR szError[64] = {0};
                wsprintf( szError, TEXT("Start(): %d"), dwRet );
                SetDlgItemText( hWnd, ID_STATIC_STATUS, szError );
                EnableWindow( GetDlgItem( hWnd, ID_BUTTON_START ), TRUE );
                EnableWindow( GetDlgItem( hWnd, ID_EDIT_PORT    ), TRUE );
                break;
                }

            SetTimer( hWnd, 1, 1000, NULL );

            SetDlgItemText( hWnd, ID_STATIC_STATUS, TEXT("Started") );

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_STOP ), TRUE );

            g_fStarted = TRUE;
			break;
            }

        case ID_BUTTON_STOP:
            SetDlgItemText( hWnd, ID_STATIC_STATUS, TEXT("Stopping...") );

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_STOP ), FALSE );

            PServer.Stop();

            KillTimer( hWnd, 1 );

            SetDlgItemText( hWnd, ID_STATIC_CPUUSAGE,            NULL );
            SetDlgItemText( hWnd, ID_STATIC_WORKER_THREADS,      NULL );
            SetDlgItemText( hWnd, ID_STATIC_WORKER_THREADS_BUSY, NULL );
            SetDlgItemText( hWnd, ID_STATIC_WORKER_IOLEN,        NULL );
            SetDlgItemText( hWnd, ID_STATIC_USER_THREADS,        NULL );
            SetDlgItemText( hWnd, ID_STATIC_USER_THREADS_BUSY,   NULL );
            SetDlgItemText( hWnd, ID_STATIC_USER_IOLEN,          NULL );
            SetDlgItemText( hWnd, ID_STATIC_CONNECTIONS,         NULL );
            SetDlgItemText( hWnd, ID_STATIC_ONCONNECT,           NULL );
            SetDlgItemText( hWnd, ID_STATIC_ONDISCONNECT,        NULL );
            SetDlgItemText( hWnd, ID_STATIC_ONREMOVECONNECTION,  NULL );

            SetDlgItemText( hWnd, ID_STATIC_STATUS, TEXT("Stopped") );

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_START ), TRUE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_PORT    ), TRUE );

            g_fStarted = FALSE;
			break;

        case ID_CHECK_ONTOP:
            SetWindowPos( hWnd, IsDlgButtonChecked( hWnd, ID_CHECK_ONTOP ) ? HWND_TOPMOST : HWND_NOTOPMOST,
                          0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE );
            break;
        }
}

//###############################################################################################################//
BOOL DlgMain_OnInitDialog( HWND hWnd, HWND hWndFocus, LPARAM lParam )
{
    g_hwndMain = hWnd;

    SetDlgItemInt( hWnd, ID_EDIT_PORT, 6060, FALSE );

    return TRUE;
}

//###############################################################################################################//
void DlgMain_OnTimer( HWND hWnd, UINT id )
{
    INFORMATION info = {0};

    PServer.QueryInformation( &info );

    SetDlgItemInt( hWnd, ID_STATIC_CPUUSAGE,            info.dwCpuUsage,               FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_WORKER_THREADS,      info.dwWorkerThreadsCount,     FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_WORKER_THREADS_BUSY, info.dwWorkerThreadsBusyCount, FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_WORKER_IOLEN,        info.dwIoCompletionLen,        FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_USER_THREADS,        info.dwUserThreadsCount,       FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_USER_THREADS_BUSY,   info.dwUserThreadsBusyCount,   FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_USER_IOLEN,          info.dwUserIoCompletionLen,    FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_CONNECTIONS,         info.dwConnectionsCount,       FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_ONCONNECT,           g_lOnConnectCount,             FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_ONDISCONNECT,        g_lOnDisconnectCount,          FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_ONREMOVECONNECTION,  g_lOnRemoveConnectionCount,    FALSE );
}

//###############################################################################################################//
void DlgMain_OnClose( HWND hWnd )
{
    if( g_fStarted ){
        SendMessage( hWnd, WM_COMMAND, ID_BUTTON_STOP, NULL );
        }

	EndDialog( hWnd, 0 );
}

//###############################################################################################################//
int WINAPI Dialog_Main( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
    switch( message ){
        Dlg_Proc( hWnd, WM_INITDIALOG, DlgMain_OnInitDialog );
        Dlg_Proc( hWnd, WM_COMMAND,    DlgMain_OnCommand    );
        Dlg_Proc( hWnd, WM_TIMER,      DlgMain_OnTimer      );
        Dlg_Proc( hWnd, WM_CLOSE,      DlgMain_OnClose      );
        }

    return 0;
}

//##############################################################################################################//
int WINAPI _tWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow )
{
    DialogBox( hInstance, MAKEINTRESOURCE( IDD_DIALOG_MAIN ), NULL, Dialog_Main );

    return 0;
}

//##############################################################################################################//